/*
 * ImageKeyframing.java
 */

//TODO: disable growth zones near del. zones

import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.awt.*;
import java.util.*;

public class ImageKeyframing extends TriangleRegions {

	public ImageKeyframing(int x, int y) {
		super(x, y);
	}
	
	public int[][] loadImage(String fname) {
		int[][] d = new int[maxX][maxY];
		try {
			BufferedImage bi = ImageIO.read(new File(fname));
			Raster r = bi.getData();
			int numBands = r.getNumBands();
			if ((r.getHeight() != maxY) || (r.getWidth() != maxX))
			{
				System.out.println("image file " + fname + " is wrong size.");
				return null;
			}
			for (int i = 0; i < maxX; i++)
				for (int j = 0; j < maxY; j++)
				{
					if (r.getSample(i,j,0) == 0) {
						d[i][j] = FILLED;
					}
					else
					{
						d[i][j] = VOID;
					}
				}
			repaint();
		} catch (IOException e)
		{
			System.out.println("couldn't read image file:" + fname);
			return null;
		}
		return d;
	}
	
	public int[][] stateDelta(String fname) {
		int [][] frameB = loadImage(fname);
		
		//compute delta
		for (int i = 0;i < maxX; i++)
			for (int j = 0; j < maxY; j++)
			{
				if (isVoid(new Point(i,j)))
				{
					if (frameB[i][j] == 0) //both void
					{ frameB[i][j] = VOID; }
					else //original is void, B is not
					{ frameB[i][j] = TO_FILL;}
				}
				else
				{
					if (frameB[i][j] == 0) //original is filled, B is not
					{ frameB[i][j] = TO_DELETE; }
					else //both filled
					{ frameB[i][j] = FILLED;}
				}
			}
		return frameB;
	}
	
	public void setEdges(int[][] frameB) {
		for (int i = 0;i < maxX; i++)
			for (int j = 0; j < maxY; j++)
			{
				int ctrVal = frameB[i][j];
				if ((ctrVal == VOID) || (ctrVal == FILLED)) continue;
				for (int k = 1; k < 7; k++) //ctrVal must be grow or delete
				{
					Point p = move(new Point(i,j),k);
					int nVal = frameB[p.x][p.y];
					if ((ctrVal == TO_FILL) && (nVal == FILLED))
						frameB[i][j] = EDGE;
					if ((ctrVal == TO_DELETE) && (nVal == VOID))
						frameB[i][j] = EDGE;
				}
			}
	}
	
	public void regionsFromImage(String fname) {
		int [][] frameB = stateDelta(fname);
		
		//set up distanceField (for distance-based binding %s)
		setDistanceField(frameB);
		
		//create triangles
		createTriRegions();
		//determine region fill state
		determineFillStates(frameB);
		
		//set up inside/outside field (to allow partial fill/delete)
		targetImageArray = loadImage(fname);
		insideTestEnabled = true;
		
		//determine region normals (for smoothing gravity)
		setRegionOrigins(stateDelta(fname));
	}
	
	public void setRegionOrigins(int[][] frameB) {
		//save old state,move frameB in
		int[][] oldState = state;
		state = frameB;
		//System.out.println("setting origins");
		HashSet er = new HashSet();
		do {
			this.repaint();
			try
			{
				Thread.sleep(2000);
			}
			catch (Exception e)
			{
				System.out.println("Sleep interrupted!");
			}
			
			//find new edge regions
			setEdges(state);
			er = edgeRegions();
			//set normals of edge regions
			Iterator i = er.iterator();
			while (i.hasNext())
			{
				setOrigin((TriRegion)i.next());
			}
			//complete the current edge regions
			i = er.iterator();
			while(i.hasNext())
			{
				fillRegion((TriRegion)i.next());
			}
			//System.out.println("er:" + er.size());
		}
		while (er.size() != 0);
		
		//restore old state
		state = oldState;
		//System.out.println("done w/ origins");
	}
	
	public HashSet edgeRegions() {
		HashSet out = new HashSet();
		for (int k = 0; k < regions.size(); k++) {
			TriRegion tr = (TriRegion) regions.get(k);
			Vector pts = tr.r.pts();
			for (int i = 0; i < pts.size(); i++) {
				Point p = (Point) pts.get(i);
				if (state[p.x][p.y] == EDGE) {
					out.add(tr);
					break;
					}
			}
		}
		return out;
	}
	
	public boolean onRegionEdge(Point p) {
		if (state[p.x][p.y] != EDGE) return false;
		return true;
	}
	
	public void setOrigin(TriRegion tr) {
		//for each edge pt in region, count non-empty neighbours
		Vector pts = tr.r.pts();
		int[] counts = new int[7];
		for (int i = 0; i< 7; i++) counts[i] = 0;
		for (int i = 0; i < pts.size(); i++) {
			Point p = (Point) pts.get(i);
			if (onRegionEdge(p)) {
				for (int k = 1; k < 7; k++) {
					Point n = move(p,k);
					if ((state[n.x][n.y] == FILLED)||
						(state[n.x][n.y] == TO_DELETE))
							counts[k]++;
				}
			}
		}
		//compute 3-way weighted avg of neighbour counts
		double[] avgs = new double[7];
		for (int i = 1; i < 7; i++) {
			avgs[i] = counts[i] + 0.75*counts[shiftDir(i,1)] +
					 0.75*counts[shiftDir(i,-1)];
		}
		//find direction
		double maxCount = 0.0;
		int dir = 3;
		for (int i = 1; i < 7; i++) {
			if (avgs[i] > maxCount) {
				dir = i;
				maxCount = avgs[i];
			}
		}
		//march in direction from center to edge
		Point g = screenToArray(tr.r.ctr());
		do {
			g = move(g,dir);
		}
		while (tr.ptInside(arrayToScreen(g)));
		tr.collapseOrigin = arrayToScreen(g);
		tr.collapseDir = dir;
	}
	
	public void fillRegion(TriRegion tr) {
		Vector pts = tr.r.pts();
		for (int i = 0; i < pts.size(); i++) {
			Point p = (Point) pts.get(i);
				switch (state[p.x][p.y]) {
					case TO_FILL:
						state[p.x][p.y] = FILLED;
						break;
					case TO_DELETE:
						state[p.x][p.y] = VOID;
						break;
					case EDGE:
						state[p.x][p.y] = tr.fill ? FILLED : VOID;
						break;
				}
			}
	}
	
	public void determineFillStates(int[][] frameB) {
		Vector regionsToKill = new Vector();
		
		for (int k = 0; k < regions.size(); k++) {
			TriRegion tr = (TriRegion) regions.get(k);
			int staticCount = 0;
			int growCount = 0;
			int delCount = 0;
			Vector pts = tr.r.pts();
			for (int i = 0; i < pts.size(); i++) {
				Point p = (Point) pts.get(i);
					switch(frameB[p.x][p.y]){
						case TO_FILL:
							growCount++;
							break;
						case TO_DELETE:
							delCount++;
							break;
						default:
							staticCount++;
					}
				}
			if (growCount+delCount <= inactiveThreshold) {
				regionsToKill.add(tr);
			}
			else if (growCount > delCount) {
				tr.fill = true;
			}
			else {
				tr.fill = false;
			}
		}
		//if any regions need to be removed, kill them
		for (int i = 0; i < regionsToKill.size(); i++) {
			regions.remove(regionsToKill.get(i));
		}
	}
	
	public void createTriRegions() {
		for (int i = 0; i < maxX - triangleSize;i+= triangleSize)
			for (int j = 0; j < maxY - 2*triangleSize; j += 2*triangleSize)
			{
				int xMid = i + (int)Math.round(Math.ceil(triangleSize / 2.0)) -1;
				//System.out.println(i + " " + xMid + " " + j);
				TriRegion tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.a = arrayToScreen(i,j);
				tr.r.b = arrayToScreen(i+triangleSize,j);
				tr.r.c = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize;
				tr.triY = j/triangleSize;
				regions.add(tr);
				
				tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.c = arrayToScreen(i,j+2*triangleSize);
				tr.r.b = arrayToScreen(i+triangleSize,j+2*triangleSize);
				tr.r.a = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize;
				tr.triY = j/triangleSize + 1;
				regions.add(tr);
				
				tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.c = arrayToScreen(i+triangleSize,j+2*triangleSize);
				tr.r.b = arrayToScreen(xMid+triangleSize,j+triangleSize);
				tr.r.a = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize + 1;
				tr.triY = j/triangleSize + 1;
				regions.add(tr);
				
				tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.a = arrayToScreen(i+triangleSize,j);
				tr.r.b = arrayToScreen(xMid+triangleSize,j+triangleSize);
				tr.r.c = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize + 1;
				tr.triY = j/triangleSize;
				regions.add(tr);
			}
	}
	
	public void setDistanceField (int frameB[][]) {
		//set initial distances
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j  <maxY; j++) {
				switch(frameB[i][j]) {
					case FILLED:
						distanceField[i][j] = 0;
						break;
					default:
						distanceField[i][j] = -1; 
				}
			}
		
		//set distance field w/ worklist algorithm
		Vector wl = new Vector();
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j  <maxY; j++) {
				if (distanceField[i][j] == 0) {
					for (int k = 1; k < 7; k++) {
						Point p = move(new Point(i,j),k);
						if (!outOfBounds(p) && !wl.contains(p)) wl.add(p);
					}
				}
			}
		
		while (wl.size() > 0) {
			Point p = (Point)wl.remove(0);
			//process point, if its df value goes down, add neighbors to wl
			boolean changed = false;
			for (int k = 1; k < 7; k++) {
				Point p2 = move(p,k);
				if ((!outOfBounds(p2)) &&
				   (distanceField[p2.x][p2.y] != -1) &&
				   ((distanceField[p2.x][p2.y] + 1 < distanceField[p.x][p.y]) ||
				   		(distanceField[p.x][p.y] == -1))) {
					distanceField[p.x][p.y] = distanceField[p2.x][p2.y] + 1;
					changed = true;
				}
			}
			
			if (changed) {
				for (int k = 1; k < 7; k++) {
					Point p2 = move(p,k);
					if ((!outOfBounds(p2)) && (!wl.contains(p2)))
						wl.add(p2);
				}
			}
		}
		
		calcActiveDistanceMax();
	}
	
	int newDistance(int x, int y) {
		int ret = -1;
		Point p = new Point (x,y);
		for (int i = 1; i < 7; i++) {
			Point p2 = move(p,i);
			if (outOfBounds(p2)) continue;
			if (distanceField[p2.x][p2.y] > -1) {
				if (ret == -1) ret = distanceField[p2.x][p2.y]+1;
				else if (ret > distanceField[p2.x][p2.y]+1)
					ret = distanceField[p2.x][p2.y]+1;
			}
		}
		return ret;
	}
	
	void calcActiveDistanceMax() {
		distanceMax = 0;
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j < maxY; j++) {
				if (!isVoid(new Point(i,j)) && (distanceField[i][j] > distanceMax)) {
					distanceMax = distanceField[i][j];
				}
			}
		origDistanceMax = distanceMax;
		System.out.println("distance max " + distanceMax);
	}
	
	void setDistanceMax(int step) {
		distanceMax = origDistanceMax - (step/decayTime);
		if (distanceMax < 0) distanceMax = 0;
		//System.out.println("distance max " + distanceMax);
	}
	
	public void createHoles() {
		int numHoles = (int)Math.round(holeDensity * maxX * maxY);
		for (int i = 0; i < numHoles; i++)
		{
			Hole h = makeHole(new Point(_r.nextInt(maxX), _r.nextInt(maxY)));
			if (h != null) holes.add(h);
		}
	}
	
	public double completionPct() {
		int goodCnt = 0;
		int badCnt = 0;
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j < maxY; j++) {
				if (targetArray[i][j] == VOID) {
					if (state[i][j] == VOID) goodCnt++;
						else badCnt++;
				}
				else {
					if (state[i][j] != VOID) goodCnt++;
						else badCnt++;
				}
			}
		//System.out.println(goodCnt + " " + badCnt);
		return goodCnt/(goodCnt+badCnt*1.0) * 100.0;
	}
	
	public void step() {
		if (ticks % decayTime == 0) setDistanceMax(ticks);
		super.step();
		if ((targetArray != null) && (ticks % 30 == 0))
			System.out.println("% done: " + completionPct());
	}
	
	public void init() {
		super.init();
		if (imageName != null) {
				state = loadImage(imageName);
				createHoles();
		}
		if (targetImage != null) {
			regionsFromImage(targetImage);
			targetArray = loadImage(targetImage);
			repaint();
		}
	}
	
	public static void main(String[] args) {
		ImageKeyframing ik = new ImageKeyframing(100, 100);
		ik.radius = 2;
		ik.setSize(410, 410);
		ik.saveFrames = true;
		ik.sleepTime = 100;
		ik.imageName = "square.png";
		ik.targetImage = "tee.png";
		ik.holeDensity = 0.05f;
		ik.mainLoop();
	}
	
	int origDistanceMax;
	
	String imageName = null;
	String targetImage = null;
	int[][] targetArray = null;
	float holeDensity = 0.01f;
	int triangleSize = 15;
	int inactiveThreshold = 0; //number of active pixels in a region before it's used
	int decayTime = 15;
	
	public static final int TO_FILL = 14;
	public static final int TO_DELETE = 28;
	public static final int EDGE = 19;
}
